/*
  Licensed under the GNU Public License.
  Copyright (C) 1998 by Thomas M. Vier, Jr. All Rights Reserved.

  wipe v0.56-2a
  by Tom Vier <thomassr@erols.com> for Tortured Souls Software

  http://www.erols.com/thomassr/zero/tortured_souls.html

  wipe is free software.
  See LICENSE for more information.

  if you've got a nice mod, send me a patch!
  freely modify and redistribute as per license.
*/

#define VERSION "0.56-2a"
#define COPYRIGHT "Copyright (C) 1998 Thomas M. Vier, Jr. All Rights Reserved."

/* includes */
#include <config.h>
#include <require.h>
#include <mt19937int.h>

/* for GNU includes */
#define __USE_BSD
#define _BSD_SOURCE
#define __USE_POSIX
#define _POSIX_SOURCE

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>

#include <errno.h>

/* conditional includes are in wipe.h to same scroll time 8) */
/* so are conditional define's */
#include <wipe.h>

/* defines */

/* exit codes */
#define OK 0
#define LOCK_FAILED 32

/* i should probably use the errno codes for exit values */
#define SUCCESS 0
#define FAILED 1
#define BAD_USAGE 2
#define NO_ACCESS 8
#define NOMEM 10

/* 25k */
#define PERCENT_ENABLE_SIZE 25600

/* global vars */

const char version[] = VERSION;
const char copyright[] = COPYRIGHT;

int exit_code = 0;           /* final exit code for main() */
char *program_name;          /* what we're being called as */
pid_t pid, ppid;             /* for getprn_char() */

char prompt[4];              /* interactive prompt */

int fd;                      /* file descriptor */
int random_source = -1;      /* /dev/urandom */

unsigned char custom_byte;   /* custom overwrite byte */

char *buffer = NULL;         /* file buffer */
size_t buffer_size;          /* i hope sizeof(size_t) == sizeof(off_t) */

struct opt options;          /* wipe options */
struct stat file_inode;      /* inode data */
struct complete percent;     /* stuff for showing completion percentage */

/* functions */

void help(int code, FILE *out)
{
  fprintf(out, ""
  "Wipe v%s\n"
  "by Tom Vier <thomassr@erols.com>\n\n"

  "Usage is %s [options] [file-list]\n\n"

  "Options:       Default: %s -ZedtV -x4 -p1\n\n"

  "-h        --   help - display this screen\n"
  "-u        --   usage\n"
  "-c        --   show copyright and license\n"
  "-w        --   show warranty information\n"
  "-i        --   interactive - confirm each file\n"
  "-I        --   disable interaction\n"
  "-f        --   force file wiping and override interaction\n"
  "-r or -R  --   recursion - traverse subdirectories\n"
  "-s        --   silent - disable percentage reporting\n"
  "-v        --   force verbose - always show percentage\n"
  "-V        --   verbose - show percentage if file is > %dk\n"
  "-d and -D --   delete (-d) or keep (-D) after wiping\n"
  "-t and -T --   stream or single byte mode - for (pseudo) random passes\n"
  "-e        --   enable entropy - use /dev/urandom\n"
  "-E        --   use faster, less secure, pseudo-random bytes\n"
  "-z        --   zero-out file - single pass of zeroes\n"
  "-Z        --   perform normal wipe passes\n"
  "-x[1-32]  --   enable (p)random passes - defaults to four\n"
  "-x and -X --   perform or don't perform random passes\n"
  "-p(1-32)  --   wipe file x number of times\n"
  "-b(0-255) --   overwrite file with this value byte\n"
  "", version, program_name, program_name, (PERCENT_ENABLE_SIZE / 1024));

  exit(code);
}

void badopt(int c)
{
  fprintf(stderr, ""
	  "Error: bad option: %c\n"
	  "Type \'%s -u\' for usage.\n"
	  "", c, program_name);
  exit(BAD_USAGE);
}

void usage(int code, FILE *out)
{
  fprintf(out, ""
  "Wipe v%s by Tom Vier <thomassr@erols.com>\n\n"

  "Usage is \n"
  "%s [-ucwsihHIfFtTdDvVzZxXeErR] [-x[1-32]] [-p(1-32)] [-b(0-255)] [files]\n\n"

  "Type \'%s -h\' for extended help.\n"
  "", version, program_name, program_name);

  exit(code);
}

void show_copy(void)
{
  printf(""
  "Wipe v%s by Tom Vier <thomassr@erols.com>\n"
  "%s\n\n"

  "Licensed under the GNU Public License.\n"
  "wipe comes with ABSOLUTELY NO WARRANTY; for details type \"%s -w\".\n"
  "", version, copyright, program_name);
}

void show_war(void)
{
  printf(""
  "Wipe v%s\n"
  "%s\n\n"

  "This program is free software; you can redistribute it and/or modify\n"
  "it under the terms of the GNU General Public License as published by\n"
  "the Free Software Foundation; either version 2 of the License, or\n"
  "(at your option) any later version.\n\n"

  "This program is distributed in the hope that it will be useful,\n"
  "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
  "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
  "GNU General Public License for more details.\n\n"

  "You should have received a copy of the GNU General Public License\n"
  "along with this program; if not, write to the Free Software\n"
  "Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\n"
  "", version, copyright);
}

/*
  percent_init -- initialize percent data
*/

void percent_init(void)
{
  int total_passes = 27;

  if (options.random)
    /* two random loops per wipe loop, so multiply it by 2 */
    total_passes += (options.random_loop + 1) * 2;

  if (options.verbose == 1 && buffer_size >= PERCENT_ENABLE_SIZE)
    {
      percent.display = 1; /* enable percentage reporting */
      percent.complete = 0; /* initialize */
    }

  if (options.verbose == 2) /* force percentage reporting */
    {
      percent.display = 1; percent.complete = 0;
    }

  percent.increment = (100.0 / total_passes) / (options.wipe_multiply + 1);
}

/*
  percent_update -- update and display completion
*/

inline void percent_update(void)
{
#ifdef PAUSE_AFTER_WIPE_PASS
  /* this is a better place, since percent_update is always called */
  printf("Hit a key:");
  fgetc(stdin);
#endif

  if (!percent.display) return; /* display disabled */
  if (!percent.complete) printf("\r    \r0%%"); /* clear "100%" */

  percent.complete += percent.increment;
  printf("\r%d%%", (int) (percent.complete)); fflush(stdout); /* display */
}

/*
  getrn_char -- return a random character
*/

char getrn_char(void)
{
  if (random_source > 0)
    return(gettrn_char());
  else
    return(getprn_char());
}

/*
  gettrn_char -- return true random character
*/

char gettrn_char(void)
{
  char entropy; /* a yummy byte of entropy */

#ifdef RAND_TEST
  if (random_source < 0)
    {
      fprintf("\rError: gettrn_char called while random_source not available\n");
      abort();
    }
#endif

  read(random_source, &entropy, 1); /* mmmm....entropy */

#ifdef RAND_TEST
  printf("/rentropy == %x\n", entropy);
#endif

  return(entropy);
}

/*
  getseed -- return an unsigned long int
             suitable for use as a seed

     getseed() uses the time, so you'll get a very regular pattern.
     we try to make it more irregular, based on the number of iterations,
     and the original file size
*/

unsigned long int getseed(void)
{
  struct timeval tv;

  gettimeofday(&tv, NULL);
  return((tv.tv_sec | tv.tv_usec) + (pid ^ ppid));
}

/*
  getprn_char -- return pseudo-random character
*/

char getprn_char(void)
{
#ifdef USE_RAND
# ifdef RAND_TEST
  unsigned long int seed;

  seed = getseed();
  srand(seed);

  printf("/rseed == %x\n", seed);
# else
  srand(getseed());
# endif

  return((char) rand());
#else
  unsigned long int seed = 0;

  /* seed must != 0 */
  while (seed == 0)
    seed = getseed();

# ifdef RAND_TEST
  printf("/rseed == %x\n", seed);
# endif

  sgenrand(seed);
  return((char) genrand());
#endif
}

/*
  getprn_char_stream -- return pseudo-random character for stream passes
                        much faster than using mt19937 for every byte.
			btw, this is very slow and not very secure at all.
			even using /dev/urandom straight is faster.
			that's why there's a warning in main().

  getprn_char_stream_init -- initializes prn_stream_save
*/

void getprn_char_stream_init(void)
{
  static unsigned long int prn_stream_save;

  /* this should be good enough to get us started */
  prn_stream_save = getrn_char();
  prn_stream_save += prn_stream_save % getrn_char();
}

char getprn_char_stream()
{
  static unsigned long int prn_stream_save, prn_stream_i;

  prn_stream_save ^= getseed(); /* xor the current value w/ a good seed */
  prn_stream_save += file_inode.st_size % prn_stream_i++;

  return((char) prn_stream_save);
}

/*
  prandom_pass -- fills file with a pseudo-random byte

  int passes   -- number of passes to perform
*/

void prandom_pass(int passes)
{
  while (passes--)
    write_pass(getprn_char());
}

/*
  prandom_stream_pass -- fills file with pseudo-random byte stream

  int passes          -- number of passes to perform
*/

void prandom_stream_pass(int passes)
{
  size_t offset = 0;

  getprn_char_stream_init();

  while (passes--)
    {
      while (offset < buffer_size)
	memset(buffer + offset++, getprn_char_stream(), 1);
      offset = 0;
      write(fd, buffer, buffer_size);
#ifdef SYNC
      fdatasync(fd);
#endif
      lseek(fd, 0, SEEK_SET);
      percent_update();
    }
}

/*
  random_pass -- fills file with random byte

  int passes  -- number of passes to perform
*/

void random_pass(int passes)
{
  while (passes--)
    write_pass(gettrn_char());
}

/*
  random_stream_pass -- fills file with random bytes

  int passes         -- number of passes to perform
*/

void random_stream_pass(int passes)
{
#ifdef DEBUG
  /* sanity check */
  if (random_source < 0)
    fprintf(stderr, 
	    "\rError: called random_pass(%d) while "
	    "random_source == %d",
	    passes, random_source);
#endif

  while (passes--)
    {
      read(random_source, buffer, buffer_size);
      write(fd, buffer, buffer_size);

#ifdef SYNC
      fdatasync(fd); /* sync data to disk */
#endif
      lseek(fd, 0, SEEK_SET); /* reset file offset */
      percent_update(); /* update completion percentage */
    }
}

/*
  write_pass -- fill file with given byte

  char wipe_byte -- byte to fill file with
*/

void write_pass(const unsigned char wipe_byte)
{
  memset(buffer, wipe_byte, buffer_size);
  write(fd, buffer, buffer_size);

#ifdef SYNC
  fdatasync(fd); /* sync data to disk */
#endif

  lseek(fd, 0, SEEK_SET); /* reset read/write-offset to 0 */
  percent_update();
}

/*
  write3_pass            -- over writes file with given bytes

  char wipe_byte1        -- first byte
  char wipe_byte2        -- second byte
  char wipe_byte3        -- third byte
*/

void write3_pass(const unsigned char wipe_byte1, const unsigned char wipe_byte2, const unsigned char wipe_byte3)
{
  size_t offset = 0;
  unsigned char wipe_stream[3];

  wipe_stream[0] = wipe_byte1;
  wipe_stream[1] = wipe_byte2;
  wipe_stream[2] = wipe_byte3;

  while (offset < buffer_size)
    {memcpy((buffer + offset), &wipe_stream, 3); offset += 3;}

  write(fd, buffer, buffer_size);

#ifdef SYNC
  fdatasync(fd); /* sync data to disk */
#endif

  lseek(fd, 0, SEEK_SET); /* reset */
  percent_update();
}

/*
  lock -- create as good a lock as possible

  char filename[]  --  for interaction
*/

#ifdef LOCKABLE
int lock(char filename[])
{
# ifndef HAVE_FCNTL
#  ifdef HAVE_LOCKF
  off_t lock_size;
#  endif
# endif

# ifdef HAVE_FCNTL
  struct flock file_lock;

  file_lock.l_type = F_WRLCK;
  file_lock.l_whence = 0;
  file_lock.l_start = 0;
  file_lock.l_len = 0;
# endif

# ifdef LOCK_TEST
  printf("attempting lock on %s...", filename);
# endif

# ifdef HAVE_FCHMOD
  /* try for a mandatory lock */
  if (fchmod(fd, 02600) != 0)
    {
#  ifdef LOCK_TEST
      fprintf(stderr, "mandatory lock failed\ntrying for advisory...");
#  else
    }
#  endif
# endif

# ifdef HAVE_FCNTL
  if (fcntl(fd, F_SETLKW, &file_lock) == 0)
    {
#  ifdef LOCK_TEST
      printf("got it\n");
#  endif
      return(0); /* lock successful */
    }

# elif HAVE_LOCKF
  lseek(fd, 0, SEEK_SET);
  lock_size = file_inode.st_size;
  while (lock_size % file_inode.st_blksize) ++lock_size;
  while (lock_size % 3) ++lock_size;

  if (lockf(fd, F_LOCK, lock_size) == 0)
    {
#  ifdef LOCK_TEST
      printf("got it\n");
#  endif
      return(0); /* lock successful */
    }

# elif HAVE_FLOCK
  if (flock(fd, LOCK_EX) == 0)
    {
#  ifdef LOCK_TEST
      printf("got it\n");
#  endif
      return(0); /* lock successful */
    }
# endif

  /* lock failed */
  if (options.force && !options.verbose)
    return(OK);

  memset(prompt, 0, sizeof(prompt)); /* clear prompt */
  while (prompt[0] != 'y' && prompt[0] != 'n')
    {
      printf("\r%s: lock failed: \'%s\'. Wipe anyway? ", program_name, filename);
      fgets(prompt, sizeof(prompt), stdin);

      if (prompt[0] == 'y') return(OK);
      if (prompt[0] == 'n') return(LOCK_FAILED);
    }

  return(0);
}
#endif

/*
  zero -- zeroes out a file. if options.custom, fills w/ custom byte

  char filename        --  path for file to be zeroed
*/

int zero(char filename[])
{
  if ((fd = open(filename, O_WRONLY)) < 0)
    {
      fprintf(stderr, "\rError: Cannot open file: %s\n", filename);
      exit_code = NO_ACCESS; return(NO_ACCESS);
    }

#ifndef SYNC
# ifndef IRIX
#  ifdef HAVE_FCNTL
  fcntl(fd, F_SETFL, O_SYNC);
#  endif
# endif
#endif

  if ((buffer = malloc(file_inode.st_size)) == NULL)
    {
      fprintf(stderr, "\rError: out of memory: file buffer: %s\n", filename);
      close(fd); exit_code = NOMEM; return(NOMEM);
    }

#ifdef LOCKABLE
  /* lock file */
  if (lock(filename))
    {
      close(fd); free(buffer); buffer = NULL;
      return(LOCK_FAILED);
    }
#endif

  if (options.custom)
    {
#ifdef DEBUG
      printf("\rusing custom byte: %d\n", custom_byte);
#endif
      memset(buffer, custom_byte, file_inode.st_size);
    }
  else
    memset(buffer, 0, file_inode.st_size);

  write(fd, buffer, file_inode.st_size);

#ifdef SYNC
  fdatasync(fd); /* sync data to disk */
#endif

  if (options.delete)
    {
#ifndef DEBUG
# ifdef HAVE_FTRUNCATE
      ftruncate(fd, 0);
# endif
#endif

#ifndef DEBUG
      if (unlink(filename) != 0)
	{
	  fprintf(stderr, 
	    "\07\rError: Unable to unlink %s: file wiped, but not deleted!\n", 
		  filename);
	  exit_code = 1; return(1);
	}
#endif
    }

  close(fd); free(buffer); buffer = NULL; return(0);
}

/*
  wipe -- runs wipe passes on a given file

  char filename -- path for file to be wiped
*/

int wipe(char filename[])
{
  int i; /* for write pass 0x11 increments */
  int wipe_loop;

#ifdef FILE_TEST
  printf("filename: %s\n", filename); return(0);
#endif

  if (options.zero) return(zero(filename)); /* cute, eh? */
  if (options.custom) return(zero(filename)); /* zero will check for custom */

  /* open unbuffered, write only */
  if ((fd = open(filename, O_WRONLY)) < 0)
    {
      fprintf(stderr, "\rError: Cannot open file: %s\n", filename);
      exit_code = NO_ACCESS; return(NO_ACCESS);
    }

#ifndef SYNC
# ifndef IRIX
#  ifdef HAVE_FCNTL
  fcntl(fd, F_SETFL, O_SYNC);
#  endif
# endif
#endif

#ifdef LOCKABLE
  /* lock file */
  if (lock(filename))
    {
      if (random_source) close(random_source);
      close(fd); return(LOCK_FAILED);
    }
#endif

  /*
     increase buffer size till we have enough space for 3 byte passes,
     cuz seg faults suck
  */
  buffer_size = file_inode.st_size; /* init */

  /* make sure we cover all allocated blocks, first */
  while (buffer_size % file_inode.st_blksize) ++buffer_size;
  while (buffer_size % 3) ++buffer_size; /* beast algorithm */

  if ((buffer = malloc(buffer_size)) == NULL)
    {
      fprintf(stderr,
	  "\rError: not enough memory for file buffer: %s\n", filename);
      exit_code = NOMEM; return(NOMEM);
    }

  percent_init();
  wipe_loop = options.wipe_multiply + 1;

  while (wipe_loop-- > 0)
    {
      /*
	comment format: 
	pass number -- binary pattern -- target encoding scheme
      */

      if (options.random)
	{
	  /* run four random passes */
	  if (options.stream)
	    {
	      if (options.true_random)
		random_stream_pass(options.random_loop + 1);

	      else prandom_stream_pass(options.random_loop + 1);
	    }
	  else
	    {
	      if (options.true_random)
		random_pass(options.random_loop + 1);

	      else prandom_pass(options.random_loop + 1);
	    }
	}

      /* fifth pass -- 01 -- RLL (1,7) and MFM */
      write_pass(0x55);

      /* sixth pass -- 10 -- same */
      write_pass(0xaa);

      /* seventh pass -- 10010010 01001001 00100100 -- RLL (2,7) and MFM */
      write3_pass(0x92, 0x49, 0x24);

      /* eighth pass -- 01001001 00100100 10010010 -- same */
      write3_pass(0x49, 0x24, 0x92);

      /* ninth pass -- 00100100 10010010 01001001 -- same */
      write3_pass(0x24, 0x92, 0x49);

      /* tenth pass -- start 0x11 increment passes */
      for (i = 0x00; i <= 0xff; i += 0x11) write_pass(i);

      /* 26 -- RLL(2,7) and MFM passes, again */
      write3_pass(0x92, 0x49, 0x24);
      write3_pass(0x49, 0x24, 0x92);
      write3_pass(0x24, 0x92, 0x49);

      /* 29 -- 01101101 10110110 11011011 -- RLL(2,7) */
      write3_pass(0x6d, 0xb6, 0xdb);
      /* 10110110 11011011 01101101 */
      write3_pass(0xb6, 0xdb, 0x6d);
      /* 11011011 01101101 10110110 */
      write3_pass(0xdb, 0x6d, 0xb6);

      if (options.random)
	{
	  /* run four random passes */
	  if (options.stream)
	    {
	      if (options.true_random)
		random_stream_pass(options.random_loop + 1);

	      else prandom_stream_pass(options.random_loop + 1);
	    }
	  else
	    {
	      if (options.true_random)
		random_pass(options.random_loop + 1);

	      else prandom_pass(options.random_loop + 1);
	    }
	}
    }

  /* done with passes */

  /*
     sometimes a float isn't enough decimals to accurately hold
     percent.increment, so we explicitly print 100% completion.
     otherwise we might end up with ((int) percent.complete) < 100;
     probably 98 or 99.
  */

  if (percent.complete)
    {
      printf("\r100%%"); percent.reported = 1;
    }

  /* ok, now let's clean-up all allocated memory and destroy file references */

  free(buffer); buffer = NULL;
  if (random_source) close(random_source);

  /* destroy file references, unless this is debug build */

  if (options.delete)
    {
#ifndef DEBUG
# ifdef HAVE_FTRUNCATE
      ftruncate(fd, 0); /* set file length to zero */
# endif
#endif

#ifndef DEBUG
      if (unlink(filename) != 0)
	{
	  fprintf(stderr, 
	    "\07\rError: Unable to unlink %s: file wiped, but not deleted!\n", 
		  filename);
	  exit_code = 1; return(1);
	}
#endif
    }

#ifdef HAVE_FSYNC
  fsync(fd);
#endif

  close(fd);
  return(0);
}

/*
  drill_down -- drill down through a directory

  char pathname[]  --  directory
*/

int drill_down(char pathname[])
{
  DIR *dir;
  struct dirent *entry;

#ifdef DEBUG
  printf("checking %s\n", pathname);
#endif

  if ((dir = opendir(pathname)) == NULL)
    {
      fprintf(stderr, "\rError: cannot open directory: %s\n", pathname);
      exit_code = NO_ACCESS; return(NO_ACCESS);
    }

  if (chdir(pathname) != 0)
    {
      fprintf(stderr, "\rError: cannot enter directory: %s\n", pathname);
      closedir(dir); exit_code = NO_ACCESS; return(NO_ACCESS);
    }

  readdir(dir);  /* skip "."  */
  readdir(dir);  /* skip ".." */

  while ((entry = readdir(dir)) != NULL) check_file(entry->d_name);

  if (chdir("..") != 0)
    {
      fprintf(stderr, "\rError: cannot enter directory: ..\n"
	              "\nWe're trapped!\nterminating process...\n");
      exit(NO_ACCESS);
    }

  closedir(dir);

#ifdef DEBUG
  if (options.delete)
    printf("removing: %s\n", pathname);
#endif

#ifndef DEBUG
  if (options.delete)
    {
      if (rmdir(pathname) != 0)
	{
	  fprintf(stderr, "\07\rError: unable to remove directory: %s\n", pathname);
	  exit_code = 1; return(1);
	}
    }
#endif

  return(0);
}

/*
  check_file -- check access and wipe(), if ok

  char pathname[]  --  file or directory
*/

int check_file(char pathname[])
{
  /* see if pathname exists */
  if (access(pathname, F_OK) != 0)
    {
      if (options.force) return(0);
      fprintf(stderr, "\rError: No such file or directory: %s\n", pathname);
      exit_code = NO_ACCESS; return(NO_ACCESS);
    }

  /* get file inode */
  if (stat(pathname, &file_inode) != 0)
    {
      fprintf(stderr, "Error: Cannot read inode: %s\n", pathname);
      exit_code = NO_ACCESS; return(NO_ACCESS);
    }

  /* see if it's a directory */
  if (S_ISDIR(file_inode.st_mode))
    {
      if (options.recursion)
	{
	  /*  force overrides interaction  */
	  if (options.interactive && !options.force)
	    {
	      memset(prompt, 0, sizeof(prompt)); /* clear prompt */
	      while (prompt[0] != 'y' && prompt[0] != 'n')
		{
		  printf("\r%s: destroy files in \'%s\'? ",
			 program_name, pathname);
		  fgets(prompt, sizeof(prompt), stdin);

		  if (prompt[0] == 'y')
		    {
		      drill_down(pathname); return(0);
		    }
		  if (prompt[0] == 'n') return(0); /* skip to next file */

		  /* else, incorrect response -- re-prompt */
		}
	    }

	  if (options.force)
	    {
	      drill_down(pathname); return(0);
	    }

          /* we want to respect file modes, so explicitly test permissions */
	  else
	    {
	      if (access(pathname, R_OK|W_OK|X_OK) == 0)
		{
		  drill_down(pathname); return(0);
		}
	      else
		{
		  memset(prompt, 0, sizeof(prompt)); /* clear prompt */
		  while (prompt[0] != 'y' && prompt[0] != 'n')
		    {
		      stat(pathname, &file_inode);

		      printf("\r%s: destroy \'%s\', overriding mode %04o?", 
		      program_name, pathname, file_inode.st_mode & 07777);

		      fgets(prompt, sizeof(prompt), stdin);

		      if (prompt[0] == 'y')
			{
			  drill_down(pathname); return(0);
			}
		      if (prompt[0] == 'n') return(0); /* skip to next file */

		      /* else, invalid response, re-prompt */
		    }
		}
	    }
	}
      else
	{
	  /* if !options.recursion */
	  fprintf(stderr, "%s: %s is a directory -- skipping\n",
		  program_name, pathname);
	  return(0);
	}
    }

  /*  force overrides interaction  */
  if (options.interactive && !options.force)
    {
      memset(prompt, 0, sizeof(prompt)); /* clear prompt */
      while (prompt[0] != 'y' && prompt[0] != 'n')
	{
	  printf("\r%s: destroy \'%s\'? ", program_name, pathname);
	  fgets(prompt, sizeof(prompt), stdin);

	  if (prompt[0] == 'y')
	    {
	      wipe(pathname); return(0);
	    }
	  if (prompt[0] == 'n') return(0); /* skip to next file */

	  /* else, incorrect response -- re-prompt */
	}
    }

  if (options.force)
    {
      wipe(pathname); return(0);
    }

  /* we want to respect file modes, so explicitly test permissions */

  else
    {
      if (access(pathname, W_OK) == 0)
	{
	  wipe(pathname); return(0);
	}
      else
	{
	  memset(prompt, 0, sizeof(prompt)); /* clear prompt */
	  while (prompt[0] != 'y' && prompt[0] != 'n')
	    {
	      printf("\r%s: destroy \'%s\', overriding mode %04o? ", 
		     program_name, pathname, file_inode.st_mode & 07777);

	      fgets(prompt, sizeof(prompt), stdin);

	      if (prompt[0] == 'y')
		{
		  wipe(pathname); return(0);
		}
	      if (prompt[0] == 'n') return(0); /* skip to next file */

	      /* else, invalid response, re-prompt */
	    }
	}
    }

  return(0);
}

/* yeah! it's main()! */

int main(int argc, char **argv)
{
  int opt;                        /* option character                      */
  int tmp;                        /* to check input range                  */
  extern int optopt;              /* getopt() stuff                        */
  extern char *optarg;            /* getopt() stuff                        */
  extern int optind, opterr;      /* getopt() stuff                        */

  opterr = 0;                     /* we'll handle bad options              */
  program_name = argv[0];         /* what we're called as                  */

  custom_byte = 0x00;
  pid = getpid(); ppid = getppid();

  /* set defaults */
  percent.display           = 0;  /* initialize                             */
  percent.reported          = 0;  /* initialize                             */
  options.true_random       = 1;  /* use /dev/urandom                       */
  options.verbose           = 1;  /* show percent if >= PERCENT_ENABLE_SIZE */
  options.recursion         = 0;  /* do not traverse directories            */
  options.zero              = 0;  /* don't just zero-out the file           */
  options.force             = 0;  /* respect file permissions               */
  options.delete            = 1;  /* remove targets                         */
  options.custom            = 0;  /* don't use a custom byte                */
  options.random            = 1;  /* perform random passes                  */
  options.stream            = 1;  /* use a (pseudo) random byte stream      */
  options.interactive       = 0;  /* don't confirm each file                */
  options.random_loop   = 4 - 1;  /* default to four (p)random passes       */
  options.wipe_multiply     = 0;  /* perform wipe loop once                 */

  memset(prompt, 0, sizeof(prompt)); /* clear prompt */

  while ((opt = getopt(argc, argv, "p:b:x::XucwsihHIfFtTdDvVzZeErR")) != -1)
    {
      switch(opt)
	{
	case 'p': /* wipe multiply */
	  tmp = atoi(optarg);

	  if (tmp < 1)
	    {
	      fprintf(stderr, "Error: Bad option: wipe multiply < 1\n");
	      exit(BAD_USAGE);
	    }

	  if (tmp > 32)
	    {
	      fprintf(stderr, "Error: Bad option: wipe multiply > 32\n");
	      exit(BAD_USAGE);
	    }

	  options.wipe_multiply = tmp - 1;
	  break;

	case 'b': /* overwrite file with byte */
	  sscanf(optarg, "%i", &tmp); /* get value */

	  if (tmp < 0)
	    {
	      fprintf(stderr, "Error: Bad option: wipe byte < 0\n");
	      exit(BAD_USAGE);
	    }

	  if (tmp > 255)
	    {
	      fprintf(stderr, "Error: Bad option: wipe byte > 255\n");
	      exit(BAD_USAGE);
	    }

	  options.custom = 1; custom_byte = tmp;
	  break;

	case 'x':  /* perform random passes */
	  if (optarg == 0)
	    options.random = 1;
	  else
	    {
	      tmp = atoi(optarg);

	      if (tmp < 0)
		{
		  fprintf(stderr, "Error: Bad option: random loop < 0\n");
		  exit(BAD_USAGE);
		}

	      if (tmp > 32)
		{
		  fprintf(stderr, "Error: Bad option: random loop > 32\n");
		  exit(BAD_USAGE);
		}

	      if (tmp == 0)
		options.random = 0;
	      else
		options.random_loop = tmp - 1;
	    }
	  break;

	case 'X':  /* don't perform random passes */
	  options.random = 0;
	  break;

	case 'e':  /* enable entropy -- true random */
	  options.true_random = 1; /* use /dev/urandom */
	  break;

	case 'E':  /* faster -- pseudo-random */
	  options.true_random = 0;
	  break;

	case 'r':  /* recursion */
	case 'R':  /* some people are used to '-R' */
	  options.recursion = 1; /* enable recursion */
	  break;

	case 'i':  /* interactive -- disables force */
	  options.force = 0;
	  options.interactive = 1;
	  break;

	case 'I':  /* non-interactive */
	  options.interactive = 0;
	  break;

	case 'f':  /* force -- ignore permissions and override interaction */
	  options.force = 1;
	  break;

	case 'F':  /* disable force */
	  options.force = 0;
	  break;

	case 't':  /* stream mode */
	  options.stream = 1;
	  break;

	case 'T':  /* single byte mode */
	  options.stream = 0;
	  break;

	case 'd':  /* delete targets */
	  options.delete = 1;
	  break;

	case 'D':  /* don't remove targets */
	  options.delete = 0;
	  break;

	case 'c':  /* copyright */
	  show_copy(); exit(0);
	  break;

	case 'w':  /* warranty */
	  show_war(); exit(0);
	  break;

	case 'u':  /* usage */
	  usage(0, stdout);
	  break;

	case 'h':  /* help */
	case 'H':
	  help(0, stdout);
	  break;

	case 'v':  /* force verbose */
	  options.verbose = 2;
	  break;

	case 'V':  /* verbose */
	  options.verbose = 1;
	  break;

	case 's':  /* silent */
	  options.verbose = 0;
	  break;

	case 'z':  /* zero */
	  options.zero = 1;
	  break;

	case 'Z':  /* don't just zero */
	  options.zero = 0;
	  break;

	default: /* bad option */
	  badopt(optopt);
	  break;
	}
    }

#ifdef OPTION_TEST
  printf("options are:\n");
  printf("true_random       = %d\n", options.true_random);
  printf("verbose           = %d\n", options.verbose);
  printf("recursion         = %d\n", options.recursion);
  printf("zero              = %d\n", options.zero);
  printf("force             = %d\n", options.force);
  printf("delete            = %d\n", options.delete);
  printf("custom            = %d\n", options.custom);
  printf("random            = %d\n", options.random);
  printf("stream            = %d\n", options.stream);
  printf("interactive       = %d\n", options.interactive);
  printf("random_loop       = %d\n", options.random_loop + 1);
  printf("wipe_multiply     = %d\n\n", options.wipe_multiply + 1);
#endif

#ifdef OPTION_TEST_ONLY
  exit(0);
#endif

#ifdef FILE_TEST
  printf("getopt() parsed %d args\n", optind - 1);
#endif

  if (optind == argc)
    {
      fprintf(stderr, "Type \'%s -u\' for usage.\n"
	              "You must specify at least one file to wipe.\n",
	      program_name);
      exit(BAD_USAGE);
    }

  if ((random_source = open("/dev/urandom", O_RDONLY)) < 0)
    if (options.true_random)
      {
	if (!options.interactive)
	  {
	    fprintf(stderr, "\rError: Cannot open file: "
		    "/dev/urandom: random data stream not available.\n");
	    exit(NO_ACCESS);
	  }
	else
	  {
	    memset(prompt, 0, sizeof(prompt)); /* clear prompt */
	    while (prompt[0] != 'y' && prompt[0] != 'n')
	      {
		printf("\r%s: can't open /dev/urandom: "
		       "use pseudo-random byte passes instead? ",
		       program_name);

		fgets(prompt, sizeof(prompt), stdin);

		if (prompt[0] == 'y')
		  {random_source = -1; options.true_random = 0;}
		if (prompt[0] == 'n') exit(NO_ACCESS);
		/* else, invalid response, re-prompt */
	      }
	  }
      }
    else
      {
	if ((random_source = open("/dev/random", O_RDONLY)) < 0)
	  random_source = -1;
      }

  if (options.random)
    if (random_source > 0)
      if (options.stream & !options.true_random)
	fprintf(stderr,
	      "%s: warning: pseudo-random streams are very slow and insecure. "
	      "Use true-random streams, if possible.\n", program_name);

  /* check access and wipe if ok */
  while (optind < argc) check_file(argv[optind++]);

#ifdef FILE_TEST
  printf("\n");
#endif

  /* change to \n if you wanna see the '100%' after wipe's done */
  if (percent.reported) printf("\r");

  return(exit_code);
}
